/*
*         OpenPBS (Portable Batch System) v2.3 Software License
*
* Copyright (c) 1999-2000 Veridian Information Solutions, Inc.
* All rights reserved.
*
* ---------------------------------------------------------------------------
* For a license to use or redistribute the OpenPBS software under conditions
* other than those described below, or to purchase support for this software,
* please contact Veridian Systems, PBS Products Department ("Licensor") at:
*
*    www.OpenPBS.org  +1 650 967-4675                  sales@OpenPBS.org
*                        877 902-4PBS (US toll-free)
* ---------------------------------------------------------------------------
*
* This license covers use of the OpenPBS v2.3 software (the "Software") at
* your site or location, and, for certain users, redistribution of the
* Software to other sites and locations.  Use and redistribution of
* OpenPBS v2.3 in source and binary forms, with or without modification,
* are permitted provided that all of the following conditions are met.
* After December 31, 2001, only conditions 3-6 must be met:
*
* 1. Commercial and/or non-commercial use of the Software is permitted
*    provided a current software registration is on file at www.OpenPBS.org.
*    If use of this software contributes to a publication, product, or
*    service, proper attribution must be given; see www.OpenPBS.org/credit.html
*
* 2. Redistribution in any form is only permitted for non-commercial,
*    non-profit purposes.  There can be no charge for the Software or any
*    software incorporating the Software.  Further, there can be no
*    expectation of revenue generated as a consequence of redistributing
*    the Software.
*
* 3. Any Redistribution of source code must retain the above copyright notice
*    and the acknowledgment contained in paragraph 6, this list of conditions
*    and the disclaimer contained in paragraph 7.
*
* 4. Any Redistribution in binary form must reproduce the above copyright
*    notice and the acknowledgment contained in paragraph 6, this list of
*    conditions and the disclaimer contained in paragraph 7 in the
*    documentation and/or other materials provided with the distribution.
*
* 5. Redistributions in any form must be accompanied by information on how to
*    obtain complete source code for the OpenPBS software and any
*    modifications and/or additions to the OpenPBS software.  The source code
*    must either be included in the distribution or be available for no more
*    than the cost of distribution plus a nominal fee, and all modifications
*    and additions to the Software must be freely redistributable by any party
*    (including Licensor) without restriction.
*
* 6. All advertising materials mentioning features or use of the Software must
*    display the following acknowledgment:
*
*     "This product includes software developed by NASA Ames Research Center,
*     Lawrence Livermore National Laboratory, and Veridian Information
*     Solutions, Inc.
*     Visit www.OpenPBS.org for OpenPBS software support,
*     products, and information."
*
* 7. DISCLAIMER OF WARRANTY
*
* THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND. ANY EXPRESS
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT
* ARE EXPRESSLY DISCLAIMED.
*
* IN NO EVENT SHALL VERIDIAN CORPORATION, ITS AFFILIATED COMPANIES, OR THE
* U.S. GOVERNMENT OR ANY OF ITS AGENCIES BE LIABLE FOR ANY DIRECT OR INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* This license will be governed by the laws of the Commonwealth of Virginia,
* without reference to its choice of law rules.
*/
/* pbs_alterjob.c

 Send the Alter Job request to the server --
 really an instance of the "manager" request.
*/

#include <pbs_config.h>   /* the master config generated by configure */

#include <stdio.h>
#include <stdlib.h>
#include "libpbs.h"

/* returns pbs_errno */

int pbs_alterjob_asyncflag(

  int           c,       /* I */
  char         *jobid,   /* I */
  struct attrl *attrib,  /* I */
  char         *extend,  /* I */
  int           asyncFlag)  /* I */

  {

  struct attropl *ap = (struct attropl *)NULL;

  struct attropl *ap1 = (struct attropl *)NULL;

  int i;

  if ((c < 0) || (jobid == NULL) || (*jobid == '\0'))
    {
    pbs_errno = PBSE_IVALREQ;

    return(pbs_errno);
    }

  /* copy the attrl to an attropl */

  while (attrib != NULL)
    {
    if (ap == NULL)
      {
      ap1 = ap = MH(struct attropl);
      }
    else
      {
      ap->next = MH(struct attropl);
      ap = ap->next;
      }

    if (ap == NULL)
      {
      /* FAILURE */

      pbs_errno = PBSE_SYSTEM;

      return(-1);
      }

    ap->name = attrib->name;

    ap->resource = attrib->resource;
    ap->value = attrib->value;

    /* attributes now need an op specified. Within TORQUE this defaults
     * to SET, ensuring the same functionality as before */

    ap->op = attrib->op;
    ap->next = (struct attropl *)NULL;
    attrib = attrib->next;
    }  /* END while (attrib != NULL) */

  i = PBSD_manager(
        c,
        asyncFlag ? PBS_BATCH_AsyModifyJob : PBS_BATCH_ModifyJob,
        MGR_CMD_SET,
        MGR_OBJ_JOB,
        jobid,
        ap1,
        extend);

  /* free up the attropl we just created */

  while (ap1 != NULL)
    {
    ap = ap1->next;

    free(ap1);

    ap1 = ap;
    }

  return(i);
  }  /* END pbs_alterjob_asyncflag() */

int pbs_alterjob_async(

  int           c,       /* I */
  char         *jobid,   /* I */
  struct attrl *attrib,  /* I */
  char         *extend)  /* I */

  {
  return(pbs_alterjob_asyncflag(c, jobid, attrib, extend, TRUE));
  }  /* END pbs_alterjob_async() */

int pbs_alterjob(

  int           c,       /* I */
  char         *jobid,   /* I */
  struct attrl *attrib,  /* I */
  char         *extend)  /* I */

  {
  return(pbs_alterjob_asyncflag(c, jobid, attrib, extend, FALSE));
  }  /* END pbs_alterjob() */

/* END pbsD_alterjo.c */

